@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2002-2017. All Rights Reserved.
+ * Copyright Ericsson AB 2002-2018. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -949,6 +949,12 @@ static int ber_decode_value(ErlNifEnv* env, ERL_NIF_TERM *value, unsigned char *
     unsigned char *tmp_out_buff;
     ERL_NIF_TERM term = 0, curr_head = 0;
 
+    /* Recursion depth limitation, borrow a signed int: maybe_ret */
+    maybe_ret = (int) (ErlNifSInt) ((char *)value - (char *)ib_index);
+    maybe_ret = maybe_ret < 0 ? -maybe_ret : maybe_ret;
+    if (maybe_ret >= sizeof(void *) * 8192) /* 8 k pointer words */
+        return ASN1_ERROR;
+
     if (((in_buf[*ib_index]) & 0x80) == ASN1_SHORT_DEFINITE_LENGTH) {
 	len = in_buf[*ib_index];
     } else if (in_buf[*ib_index] == ASN1_INDEFINITE_LENGTH) {
@@ -993,7 +999,7 @@ static int ber_decode_value(ErlNifEnv* env, ERL_NIF_TERM *value, unsigned char *
 	while (*ib_index < end_index) {
 
 	    if ((maybe_ret = ber_decode(env, &term, in_buf, ib_index,
-		    in_buf_len)) <= ASN1_ERROR
+                   end_index )) <= ASN1_ERROR
 	    )
 		return maybe_ret;
 	    curr_head = enif_make_list_cell(env, term, curr_head);
@@ -1274,17 +1280,20 @@ static ERL_NIF_TERM encode_ber_tlv(ErlNifEnv* env, int argc,
     ERL_NIF_TERM err_code;
 
     curr = ber_new_chunk(40);
+    if (!curr) {
+        err_code = enif_make_atom(env,"oom");
+        goto err;
+    }
 
-    if ((encode_err = ber_encode(env, argv[0], &curr, &length))
-	    <= ASN1_ERROR) {
-	ber_free_chunks(curr);
+    encode_err = ber_encode(env, argv[0], &curr, &length);
+    if (encode_err <= ASN1_ERROR) {
 	err_code = enif_make_int(env, encode_err);
-	return enif_make_tuple2(env, enif_make_atom(env, "error"), err_code);
+	goto err;
     }
 
     if (!enif_alloc_binary(length, &out_binary)) {
-	ber_free_chunks(curr);
-	return enif_make_tuple2(env, enif_make_atom(env, "error"), enif_make_atom(env,"oom"));
+        err_code = enif_make_atom(env,"oom");
+        goto err;
     }
 
     top = curr;
@@ -1300,6 +1309,10 @@ static ERL_NIF_TERM encode_ber_tlv(ErlNifEnv* env, int argc,
     ber_free_chunks(top);
 
     return enif_make_binary(env, &out_binary);
+
+ err:
+    ber_free_chunks(curr);
+    return enif_make_tuple2(env, enif_make_atom(env, "error"), err_code);
 }
 
 static int is_ok_load_info(ErlNifEnv* env, ERL_NIF_TERM load_info) {
